from django.contrib.auth import authenticate, login
from django.contrib.auth import logout as auth_logout
from django.contrib.auth.mixins import LoginRequiredMixin
from django.contrib.auth.views import (
    LoginView,
    PasswordChangeDoneView,
    PasswordChangeView,
    PasswordResetCompleteView,
    PasswordResetConfirmView,
    PasswordResetDoneView,
)
from django.core.exceptions import ValidationError
from django.http import HttpResponseRedirect
from django.shortcuts import render
from django.urls import reverse_lazy
from django.utils.decorators import method_decorator
from django.utils.http import urlsafe_base64_decode
from django.utils.translation import gettext_lazy as _
from django.views.decorators.cache import never_cache
from django.views.decorators.csrf import csrf_protect
from django.views.decorators.debug import sensitive_post_parameters
from django.views.generic.edit import FormView

from phone_auth.mixins import AnonymousRequiredMixin

from . import app_settings
from .forms import (
    AddEmailForm,
    AddPhoneForm,
    PhoneEmailVerificationForm,
    PhoneLoginForm,
    PhoneLogoutForm,
    PhonePasswordResetForm,
    PhoneRegisterForm,
)
from .models import EmailAddress, PhoneNumber
from .tokens import phone_token_generator


class PhoneSignupView(AnonymousRequiredMixin, FormView):
    """Display the register form and handle user registration."""

    form_class = PhoneRegisterForm
    template_name = "phone_auth/signup.html"
    success_url = reverse_lazy("phone_auth:phone_login")

    @method_decorator(sensitive_post_parameters())
    @method_decorator(csrf_protect)
    @method_decorator(never_cache)
    def dispatch(self, *args, **kwargs):
        return super(PhoneSignupView, self).dispatch(*args, **kwargs)

    def form_valid(self, form):
        form.save()
        if form.errors:
            return render(self.request, self.template_name, context={"form": form})
        return super().form_valid(form)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        return context


class PhoneLoginView(AnonymousRequiredMixin, LoginView):
    """Display the login form and handle the login action."""

    form_class = PhoneLoginForm
    template_name = "phone_auth/login.html"

    def form_valid(self, form):
        """Security check complete. Log the user in."""

        user = authenticate(self.request, **form.cleaned_data)
        if user is not None:
            login(self.request, user)
        else:
            form.add_error("login", "Invalid Credentials")
            return render(
                self.request,
                "phone_auth/login.html",
                context={"form": form},
                status=400,
            )
        return HttpResponseRedirect(self.get_success_url())


class PhoneLogoutView(FormView):
    """Handle logout"""

    template_name = "phone_auth/logout.html"
    form_class = PhoneLogoutForm
    success_url = _(app_settings.LOGOUT_REDIRECT_URL)

    @method_decorator(csrf_protect)
    @method_decorator(never_cache)
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)

    def form_valid(self, form):
        auth_logout(self.request)
        return super().form_valid(form)


class PhonePasswordResetView(FormView):
    """Display the password reset form and handle password reset using phone/email."""

    form_class = PhonePasswordResetForm
    template_name = "phone_auth/password_reset.html"
    success_url = reverse_lazy("phone_auth:phone_password_reset_done")

    @method_decorator(csrf_protect)
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)

    def form_valid(self, form):
        form.save()
        return super().form_valid(form)


class PhonePasswordResetDoneView(PasswordResetDoneView):
    """Renders a template."""

    template_name = "phone_auth/password_reset_done.html"


class PhonePasswordConfirmView(PasswordResetConfirmView):
    """Accepts `uidb64` and `token` kwargs and validates them.

    If valid, it renders a form for the user to reset the password.
    """

    template_name = "phone_auth/password_reset_confirm.html"
    success_url = reverse_lazy("phone_auth:phone_password_reset_complete")


class PhonePasswordResetCompleteView(PasswordResetCompleteView):
    """Renders a template"""

    template_name = "phone_auth/password_reset_complete.html"


class PhoneChangePasswordView(PasswordChangeView):
    """View to change password using old password"""

    template_name = "phone_auth/change_password.html"
    success_url = reverse_lazy("phone_auth:phone_change_password_done")


class PhoneChangePasswordDoneView(PasswordChangeDoneView):
    """Renders a template"""

    template_name = "phone_auth/change_password_done.html"


class PhoneEmailVerificationView(LoginRequiredMixin, FormView):
    """
    Display all email-addresses and phone-numbers associated with user
    account with verification status on GET request.
    """

    template_name = "phone_auth/phone_and_email_verification.html"
    form_class = PhoneEmailVerificationForm
    title = None

    @method_decorator(csrf_protect)
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super(PhoneEmailVerificationView, self).get_context_data()
        context.update(
            {
                "email_addresses": self.request.user.emailaddress_set.all(),
                "phone_numbers": self.request.user.phonenumber_set.all(),
            }
        )
        if self.title is not None:
            context["title"] = self.title
        return context

    def form_valid(self, form):
        self.title = form.save(self.request.user)
        return render(
            self.request,
            template_name=self.template_name,
            context=self.get_context_data(),
        )


class PhoneEmailVerificationConfirmView(FormView):
    """Accepts `idb64` and `token` kwargs and validates them.

    If valid, it set is_verified to True of PhoneNumber/EmailAdress
    instance.
    """

    template_name = "phone_auth/phone_email_verification_confirm.html"

    # noinspection PyAttributeOutsideInit
    @method_decorator(sensitive_post_parameters())
    @method_decorator(never_cache)
    def dispatch(self, *args, **kwargs):
        assert "idb64" in kwargs and "token" in kwargs

        self.validlink = False
        email_obj, phone_obj = self.get_email_or_phone_obj(kwargs["idb64"])

        if email_obj is not None:
            user = email_obj.user
        elif phone_obj is not None:
            user = phone_obj.user
        else:
            user = None

        if user:
            is_valid_token = phone_token_generator(
                email_address_obj=email_obj, phone_number_obj=phone_obj
            ).check_token(user, kwargs["token"])

            if is_valid_token:
                if email_obj is not None:
                    email_obj.is_verified = True
                    email_obj.save()
                if phone_obj is not None:
                    phone_obj.is_verified = True
                    phone_obj.save()
                self.validlink = True

        # Display the "Verification Failed/Passed" page.
        return self.render_to_response(self.get_context_data())

    @staticmethod
    def get_email_or_phone_obj(idb64):
        email_obj = phone_obj = None
        try:
            # urlsafe_base64_decode() decodes to bytestring
            uid = urlsafe_base64_decode(idb64).decode()
            method = uid[:5]
            pk = int(uid[5:])
            if method == "email":
                email_obj = EmailAddress.objects.get(pk=pk)
            elif uid[:5] == "phone":
                phone_obj = PhoneNumber.objects.get(pk=pk)

        except (
            TypeError,
            ValueError,
            OverflowError,
            EmailAddress.DoesNotExist,
            PhoneNumber.DoesNotExist,
            ValidationError,
        ):
            pass
        return email_obj, phone_obj

    def get_context_data(self, **kwargs):
        context = {}
        if self.validlink:
            context["title"] = _("Verification successful")
        else:
            context["title"] = _("Verification failed")
        return context


class AddPhoneView(LoginRequiredMixin, FormView):
    """Add new phone"""

    template_name = "phone_auth/add_new_phone.html"
    form_class = AddPhoneForm
    success_url = reverse_lazy("phone_auth:phone_email_verification")

    @method_decorator(csrf_protect)
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)

    def form_valid(self, form):
        form.save(self.request.user)
        if form.errors:
            return render(self.request, self.template_name, context={"form": form})
        return super().form_valid(form)


class AddEmailView(LoginRequiredMixin, FormView):
    """Add new email"""

    template_name = "phone_auth/add_new_email.html"
    form_class = AddEmailForm
    success_url = reverse_lazy("phone_auth:phone_email_verification")

    @method_decorator(csrf_protect)
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)

    def form_valid(self, form):
        form.save(self.request.user)
        if form.errors:
            return render(self.request, self.template_name, context={"form": form})
        return super().form_valid(form)



from django.contrib.auth import get_user_model
from django.db import models

# noinspection PyUnresolvedReferences
from phonenumber_field.modelfields import PhoneNumberField

User = get_user_model()


class PhoneNumber(models.Model):
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    phone = PhoneNumberField(unique=True, blank=False)
    is_verified = models.BooleanField(default=False)

    def __str__(self):
        return str(self.phone)


class EmailAddress(models.Model):
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    email = models.EmailField(unique=True, blank=False)
    is_verified = models.BooleanField(default=False)

    def __str__(self):
        return self.email



import sys

from django.conf import settings

from .exceptions import AuthenticationMethodEmpty


class AppSettings:
    class AuthenticationMethod:
        USERNAME = "username"
        EMAIL = "email"
        PHONE = "phone"

    @property
    def AUTHENTICATION_METHODS(self):
        default = {
            self.AuthenticationMethod.USERNAME,
            self.AuthenticationMethod.EMAIL,
            self.AuthenticationMethod.PHONE,
        }
        auth_methods = self._setting("AUTHENTICATION_METHODS", default)
        if not auth_methods:
            raise AuthenticationMethodEmpty
        return auth_methods

    @property
    def REGISTER_USERNAME_REQUIRED(self):
        default = True
        return self._setting("REGISTER_USERNAME_REQUIRED", default)

    @property
    def REGISTER_EMAIL_REQUIRED(self):
        default = True
        return self._setting("REGISTER_EMAIL_REQUIRED", default)

    @property
    def REGISTER_FNAME_REQUIRED(self):
        default = True
        return self._setting("REGISTER_FNAME_REQUIRED", default)

    @property
    def REGISTER_LNAME_REQUIRED(self):
        default = True
        return self._setting("REGISTER_LNAME_REQUIRED", default)

    @property
    def REGISTER_CONFIRM_PASSWORD_REQUIRED(self):
        default = True
        return self._setting("REGISTER_CONFIRM_PASSWORD_REQUIRED", default)

    @property
    def LOGIN_REDIRECT_URL(self):
        default = "/accounts/profile/"
        return self._setting("LOGIN_REDIRECT_URL", default)

    @property
    def LOGOUT_REDIRECT_URL(self):
        default = "/"
        return self._setting("LOGOUT_REDIRECT_URL", default)

    @staticmethod
    def _setting(name, default):
        ret = getattr(settings, name, default)
        return ret if ret is not None else default


app_settings = AppSettings()
app_settings.__name__ = __name__
# noinspection PyTypeChecker
sys.modules[__name__] = app_settings



from django.urls import path

from .views import (
    AddEmailView,
    AddPhoneView,
    PhoneChangePasswordDoneView,
    PhoneChangePasswordView,
    PhoneEmailVerificationConfirmView,
    PhoneEmailVerificationView,
    PhoneLoginView,
    PhoneLogoutView,
    PhonePasswordConfirmView,
    PhonePasswordResetCompleteView,
    PhonePasswordResetDoneView,
    PhonePasswordResetView,
    PhoneSignupView,
)

app_name = "phone_auth"

urlpatterns = [
    path("signup/", PhoneSignupView.as_view(), name="phone_signup"),
    path("login/", PhoneLoginView.as_view(), name="phone_login"),
    path("logout/", PhoneLogoutView.as_view(), name="phone_logout"),
    path(
        "password_reset/", PhonePasswordResetView.as_view(), name="phone_password_reset"
    ),
    path(
        "password_reset_done/",
        PhonePasswordResetDoneView.as_view(),
        name="phone_password_reset_done",
    ),
    path(
        "password_reset_confirm/<uidb64>/<token>/",
        PhonePasswordConfirmView.as_view(),
        name="phone_password_reset_confirm",
    ),
    path(
        "password_reset_complete/",
        PhonePasswordResetCompleteView.as_view(),
        name="phone_password_reset_complete",
    ),
    path(
        "change_password/",
        PhoneChangePasswordView.as_view(),
        name="phone_change_password",
    ),
    path(
        "change_password_done/",
        PhoneChangePasswordDoneView.as_view(),
        name="phone_change_password_done",
    ),
    path(
        "user_verification/",
        PhoneEmailVerificationView.as_view(),
        name="phone_email_verification",
    ),
    path(
        "user_verification_confirm/<idb64>/<token>/",
        PhoneEmailVerificationConfirmView.as_view(),
        name="phone_email_verification_confirm",
    ),
    path(
        "phone/add/",
        AddPhoneView.as_view(),
        name="add_phone",
    ),
    path(
        "email/add/",
        AddEmailView.as_view(),
        name="add_email",
    ),
]



class AuthenticationMethodEmpty(Exception):
    """Exception raised if AUTHENTICATION_METHODS is set to
    empty set in settings

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message="AUTHENTICATION_METHODS can't be empty"):
        self.message = message
        super().__init__(self.message)



import string

from django.contrib.auth import REDIRECT_FIELD_NAME
from django.contrib.auth.hashers import check_password, make_password
from django.contrib.auth.models import User
from django.contrib.auth.tokens import default_token_generator
from django.core.exceptions import ValidationError
from django.http import HttpResponse
from django.test import TestCase
from django.urls import reverse
from django.utils.encoding import force_bytes
from django.utils.http import urlsafe_base64_encode
from django.views.generic import View

from phone_auth import app_settings
from phone_auth.app_settings import AuthenticationMethod
from phone_auth.decorators import (
    anonymous_required,
    verified_email_required,
    verified_phone_required,
)
from phone_auth.mixins import (
    AnonymousRequiredMixin,
    VerifiedEmailRequiredMixin,
    VerifiedPhoneRequiredMixin,
)
from phone_auth.models import EmailAddress, PhoneNumber
from phone_auth.tokens import phone_token_generator
from phone_auth.validators import validate_username


class AccountTests(TestCase):
    data = {
        "phone": "+919876543210",
        "username": "test",
        "email": "someone@example.com",
        "first_name": "first",
        "last_name": "last",
        "password": "abcd@1234",
    }

    @classmethod
    def setUpTestData(cls):
        user_data = dict(cls.data)
        phone = user_data.pop("phone")
        user_data["password"] = make_password(user_data["password"])

        cls.user = User.objects.create(**user_data)
        cls.phone_obj = PhoneNumber.objects.create(user=cls.user, phone=phone)
        cls.email_obj = EmailAddress.objects.create(
            user=cls.user, email=user_data["email"]
        )

    def test_phone_signup_view(self):

        url = reverse("phone_auth:phone_signup")
        data = {
            "phone": "+919999999999",
            "username": "test1",
            "email": "someone@register.com",
            "first_name": "first",
            "last_name": "last",
            "password": "abcd@1234",
            "confirm_password": "abcd@1234",
        }

        response = self.client.post(url, data)
        self.assertEqual(response.status_code, 302)

        user = User.objects.latest("id")
        phone = user.phonenumber_set.latest("id").phone
        email = user.emailaddress_set.latest("id").email

        self.assertEqual(phone.__str__(), data["phone"])
        self.assertEqual(email, data["email"])
        self.assertTrue(check_password(data["password"], user.password))
        self.assertEqual(user.username, data["username"])
        self.assertEqual(user.email, data["email"])
        self.assertEqual(user.first_name, data["first_name"])
        self.assertEqual(user.last_name, data["last_name"])

    def test_phone_login_view(self):
        url = reverse("phone_auth:phone_login")
        authentication_methods = app_settings.AUTHENTICATION_METHODS

        # With correct password
        for auth_method in authentication_methods:
            credentials = {
                "login": self.data[auth_method],
                "password": self.data["password"],
            }
            response = self.client.post(url, credentials)
            self.assertEqual(response.status_code, 302)
            self.assertTrue(response.wsgi_request.user.is_authenticated)

            # Logout
            self.client.logout()

        # With incorrect password
        for auth_method in authentication_methods:
            credentials = {
                "login": self.data[auth_method],
                "password": "inco@rrect0Pass",
            }
            response = self.client.post(url, credentials)
            self.assertEqual(response.status_code, 400)
            self.assertFalse(response.wsgi_request.user.is_authenticated)

    def test_phone_logout_view(self):
        # Login
        self.client.login(login=self.data["email"], password=self.data["password"])

        # Logout
        url = reverse("phone_auth:phone_logout")
        response = self.client.post(url)
        self.assertEqual(response.status_code, 302)
        self.assertFalse(response.wsgi_request.user.is_authenticated)

    def test_phone_password_reset_view(self):
        url = reverse("phone_auth:phone_password_reset")

        for method in ["phone", "email"]:
            data = {"login": self.data[method]}
            response = self.client.post(url, data)
            self.assertEqual(response.status_code, 302)

    def test_phone_password_reset_done_view(self):
        url = reverse("phone_auth:phone_password_reset_done")
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

    def test_phone_password_reset_confirm_view(self):
        # generate `uidb64` and `token`
        credentials = {
            "uidb64": urlsafe_base64_encode(force_bytes(self.user.pk)),
            "token": default_token_generator.make_token(self.user),
        }

        url = reverse("phone_auth:phone_password_reset_confirm", kwargs=credentials)
        response = self.client.get(url)
        self.assertEqual(response.status_code, 302)
        # self.assertEqual(response.url[-14:], '/set-password/')

    def test_phone_password_reset_complete_view(self):
        url = reverse("phone_auth:phone_password_reset_complete")
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

    def test_phone_change_password_view(self):
        url = reverse("phone_auth:phone_change_password")
        data = {
            "old_password": self.data.get("password"),
            "new_password": "abcd@1234",
            "confirm_password": "abcd@1234",
        }
        response = self.client.post(url, data)

        user = User.objects.get(email=self.data["email"])
        self.assertEqual(response.status_code, 302)
        self.assertTrue(check_password(data["new_password"], user.password))

    def test_phone_token_generator(self):
        # Test with email
        email_obj = self.user.emailaddress_set.all()[0]
        token_generator = phone_token_generator(
            email_address_obj=email_obj, phone_number_obj=None
        )
        token = token_generator.make_token(self.user)

        self.assertIsNotNone(token)
        self.assertTrue(token_generator.check_token(self.user, token))

        # Test with phone
        phone_obj = self.user.phonenumber_set.all()[0]
        token_generator = phone_token_generator(
            email_address_obj=None, phone_number_obj=phone_obj
        )
        token = token_generator.make_token(self.user)

        self.assertIsNotNone(token)
        self.assertTrue(token_generator.check_token(self.user, token))

    def test_phone_email_verification_view(self):
        # Login
        self.client.login(login=self.data["email"], password=self.data["password"])

        # Test
        url = reverse("phone_auth:phone_email_verification")

        # GET request
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

        # POST request (email)
        data = {"method": "email", "pk": self.email_obj.pk}
        response = self.client.post(url, data)
        self.assertEqual(response.status_code, 200)

        # POST request (phone)
        data = {"method": "phone", "pk": self.phone_obj.pk}
        response = self.client.post(url, data)
        self.assertEqual(response.status_code, 200)

    def test_phone_email_verification_confirm_view(self):
        # Email verification
        self.assertFalse(self.email_obj.is_verified)
        credentials = {
            "idb64": urlsafe_base64_encode(force_bytes(f"email{self.email_obj.pk}")),
            "token": phone_token_generator(
                email_address_obj=self.email_obj, phone_number_obj=None
            ).make_token(self.user),
        }
        url = reverse("phone_auth:phone_email_verification_confirm", kwargs=credentials)

        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertTrue(EmailAddress.objects.get(pk=self.email_obj.pk).is_verified)

        # Phone Verification
        self.assertFalse(self.phone_obj.is_verified)
        credentials = {
            "idb64": urlsafe_base64_encode(force_bytes(f"phone{self.phone_obj.pk}")),
            "token": phone_token_generator(
                email_address_obj=None, phone_number_obj=self.phone_obj
            ).make_token(self.user),
        }
        url = reverse("phone_auth:phone_email_verification_confirm", kwargs=credentials)

        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertTrue(PhoneNumber.objects.get(pk=self.phone_obj.pk).is_verified)

    def test_anonymous_required_decorator(self):
        @anonymous_required
        def tview(request):
            return HttpResponse()

        request = self.client.get("/").wsgi_request
        self.assertTrue(request.user.is_anonymous)
        response = tview(request)
        self.assertEqual(response.status_code, 200)

        # checking for logged in user
        self.client.login(login=self.data["email"], password=self.data["password"])
        request = self.client.get("/").wsgi_request
        self.assertTrue(request.user.is_authenticated)
        response = tview(request)
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            response.url, f"{app_settings.LOGIN_REDIRECT_URL}?{REDIRECT_FIELD_NAME}=/"
        )

    def test_verified_email_required_decorator(self):

        if AuthenticationMethod.EMAIL in app_settings.AUTHENTICATION_METHODS:

            @verified_email_required
            def tview(request):
                return HttpResponse()

            self.client.login(login=self.data["email"], password=self.data["password"])
            request = self.client.get("/").wsgi_request

            # check with not verified email
            response = tview(request)
            self.assertEqual(response.status_code, 302)
            self.assertEqual(
                response.url, reverse("phone_auth:phone_email_verification")
            )

            # check with verified email
            self.email_obj.is_verified = True
            self.email_obj.save()
            response = tview(request)
            self.assertEqual(response.status_code, 200)

    def test_verified_phone_required_decorator(self):

        if AuthenticationMethod.PHONE in app_settings.AUTHENTICATION_METHODS:

            @verified_phone_required
            def tview(request):
                return HttpResponse()

            self.client.login(login=self.data["phone"], password=self.data["password"])
            request = self.client.get("/").wsgi_request

            # check with not verified phone
            response = tview(request)
            self.assertEqual(response.status_code, 302)
            self.assertEqual(
                response.url, reverse("phone_auth:phone_email_verification")
            )

            # check with verified email
            self.phone_obj.is_verified = True
            self.phone_obj.save()
            response = tview(request)
            self.assertEqual(response.status_code, 200)

    def test_anonymous_required_mixin(self):
        class Tview(AnonymousRequiredMixin, View):
            def get(self, request, *args, **kwargs):
                return HttpResponse()

        request = self.client.get("/").wsgi_request
        response = Tview.as_view()(request)
        self.assertEqual(response.status_code, 200)

        # checking for logged in user
        self.client.login(login=self.data["email"], password=self.data["password"])
        request = self.client.get("/").wsgi_request
        response = Tview.as_view()(request)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, app_settings.LOGIN_REDIRECT_URL)

    def test_verified_phone_required_mixin(self):

        if AuthenticationMethod.PHONE in app_settings.AUTHENTICATION_METHODS:

            class Tview(VerifiedPhoneRequiredMixin, View):
                def get(self, request, *args, **kwargs):
                    return HttpResponse()

            self.client.login(login=self.data["phone"], password=self.data["password"])

            request = self.client.get("/").wsgi_request

            # check with not verified phone
            response = Tview.as_view()(request)
            self.assertEqual(response.status_code, 302)
            self.assertEqual(
                response.url, reverse("phone_auth:phone_email_verification")
            )

            # check with verified phone
            self.phone_obj.is_verified = True
            self.phone_obj.save()
            response = Tview.as_view()(request)
            self.assertEqual(response.status_code, 200)

    def test_verified_email_required_mixin(self):

        if AuthenticationMethod.EMAIL in app_settings.AUTHENTICATION_METHODS:

            class Tview(VerifiedEmailRequiredMixin, View):
                def get(self, request, *args, **kwargs):
                    return HttpResponse()

            self.client.login(login=self.data["email"], password=self.data["password"])
            request = self.client.get("/").wsgi_request

            # check with not verified email
            response = Tview.as_view()(request)
            self.assertEqual(response.status_code, 302)
            self.assertEqual(
                response.url, reverse("phone_auth:phone_email_verification")
            )

            # check with verified email
            self.email_obj.is_verified = True
            self.email_obj.save()
            response = Tview.as_view()(request)
            self.assertEqual(response.status_code, 200)

    def test_username_validator(self):
        self.assertIsNone(validate_username("a"))
        self.assertIsNone(validate_username("A"))
        self.assertIsNone(validate_username("a" * 150))
        self.assertIsNone(validate_username(string.digits))
        self.assertIsNone(validate_username(string.ascii_lowercase))
        self.assertIsNone(validate_username(string.ascii_uppercase))
        self.assertIsNone(validate_username("."))
        self.assertIsNone(validate_username("-"))
        self.assertIsNone(validate_username("_"))
        try:
            self.assertIsNone(validate_username(""))
            # If exception not raised, test case failed.
            self.assertTrue(False)
        except ValidationError:
            pass
        try:
            self.assertIsNone(validate_username("+"))
            self.assertTrue(False)
        except ValidationError:
            pass
        try:
            self.assertIsNone(validate_username("@"))
            self.assertTrue(False)
        except ValidationError:
            pass

    def test_add_email_view(self):
        # Login
        self.client.login(login=self.data["email"], password=self.data["password"])

        url = reverse("phone_auth:add_email")

        # test with different mail should pass
        data = {"email": "someone1@register.com"}
        response = self.client.post(url, data)
        self.assertEqual(response.status_code, 302)

        self.assertTrue(EmailAddress.objects.filter(email=data["email"]).exists())

    def test_add_phone_view(self):
        # Login
        self.client.login(login=self.data["phone"], password=self.data["password"])

        url = reverse("phone_auth:add_phone")

        # test with different phone no should pass
        data = {"phone": "+919876543211"}
        response = self.client.post(url, data)
        self.assertEqual(response.status_code, 302)

        self.assertTrue(PhoneNumber.objects.filter(phone=data["phone"]).exists())



import uuid

from django import forms
from django.contrib.auth import get_user_model, password_validation
from django.contrib.auth.hashers import make_password
from django.contrib.auth.tokens import default_token_generator
from django.core.exceptions import ValidationError
from django.db import DatabaseError, IntegrityError, transaction
from django.db.models import Q
from django.urls import reverse
from django.utils.encoding import force_bytes
from django.utils.http import urlsafe_base64_encode

# noinspection PyUnresolvedReferences
from phonenumber_field.formfields import PhoneNumberField

from phone_auth.validators import validate_username

from . import app_settings
from .models import EmailAddress, PhoneNumber
from .signals import (
    reset_password_email,
    reset_password_phone,
    verify_email,
    verify_phone,
)
from .tokens import phone_token_generator

User = get_user_model()


class PhoneRegisterForm(forms.Form):
    """Form for user registration"""

    phone = PhoneNumberField()
    username = forms.CharField(
        required=app_settings.REGISTER_USERNAME_REQUIRED, validators=[validate_username]
    )
    email = forms.EmailField(required=app_settings.REGISTER_EMAIL_REQUIRED)
    first_name = forms.CharField(required=app_settings.REGISTER_FNAME_REQUIRED)
    last_name = forms.CharField(required=app_settings.REGISTER_LNAME_REQUIRED)
    password = forms.CharField(widget=forms.PasswordInput())
    confirm_password = forms.CharField(
        widget=forms.PasswordInput(),
        required=app_settings.REGISTER_CONFIRM_PASSWORD_REQUIRED,
    )

    def clean(self):
        errors = {}
        if app_settings.REGISTER_CONFIRM_PASSWORD_REQUIRED:
            if self.cleaned_data.get("password") != self.cleaned_data.get(
                "confirm_password"
            ):
                errors["confirm_password"] = "Password didn't match"
        if self.cleaned_data.get("email", None) is not None:
            if EmailAddress.objects.filter(
                email__iexact=self.cleaned_data.get("email")
            ).exists():
                errors["email"] = "Email already exists"
        if self.cleaned_data.get("phone", None) is not None:
            if PhoneNumber.objects.filter(
                phone=self.cleaned_data.get("phone")
            ).exists():
                errors["phone"] = "Phone already exists"
        if self.cleaned_data.get("username", None) is not None:
            if User.objects.filter(
                username__exact=self.cleaned_data.get("username")
            ).exists():
                errors["username"] = "Username already exists"

        if errors:
            raise ValidationError(errors)

    def _post_clean(self):
        super()._post_clean()
        # Validate the password after self.instance is updated with form data
        # by super().
        password = self.cleaned_data.get("password")
        user_data = dict(self.cleaned_data)
        if "phone" in user_data:
            user_data.pop("phone")
        if "confirm_password" in user_data:
            user_data.pop("confirm_password")
        user_instance = User(**user_data)

        try:
            password_validation.validate_password(password, user=user_instance)
        except ValidationError as error:
            self.add_error("password", error)

    def save(self):
        try:
            phone = self.cleaned_data.get("phone", None)
            if phone is not None:
                self.cleaned_data.pop("phone")

            if "confirm_password" in self.cleaned_data:
                self.cleaned_data.pop("confirm_password")

            username = self.cleaned_data.get("username")
            if not username:
                self.cleaned_data["username"] = uuid.uuid4().hex

            email = self.cleaned_data.get("email", None)

            self.cleaned_data["password"] = make_password(self.cleaned_data["password"])

            with transaction.atomic():
                user = User.objects.create(**self.cleaned_data)
                if phone is not None:
                    PhoneNumber.objects.create(user=user, phone=phone)
                if email is not None:
                    EmailAddress.objects.create(user=user, email=email)
        except DatabaseError as e:
            if "UNIQUE constraint" in e.args[0]:
                if "phone" in e.args[0]:
                    self.add_error("phone", "Phone already exists")
                if "username" in e.args[0]:
                    self.add_error("username", "Username already exists")
                if "email" in e.args[0]:
                    self.add_error("email", "Email already exists")


class PhoneLoginForm(forms.Form):
    """Form used for user login"""

    login = forms.CharField()
    password = forms.CharField(widget=forms.PasswordInput())

    def __init__(self, request=None, *args, **kwargs):
        self.request = request
        super(PhoneLoginForm, self).__init__(*args, **kwargs)


class EmailValidationForm(forms.Form):
    """Form to validate email field"""

    email = forms.EmailField()


class PhoneValidationForm(forms.Form):
    """Form to validate phone field"""

    phone = PhoneNumberField()


class UsernameValidationForm(forms.Form):
    """Form to validate username field"""

    username = forms.CharField(validators=[validate_username])


class PhonePasswordResetForm(forms.Form):
    """Checks if the user with provided email/phone exists.

    If provided email/phone exists, send reset_password_email/reset_password_phone
    signal with user and URL (relative_path that is one-time use only link
    to reset password) arguments.
    """

    login = forms.CharField()

    @staticmethod
    def get_users_and_method(login):
        lookup_obj = Q()

        is_phone = False
        if PhoneValidationForm({"phone": login}).is_valid():
            lookup_obj |= Q(phonenumber__phone=login)
            is_phone = True

        elif EmailValidationForm({"email": login}).is_valid():
            lookup_obj |= Q(emailaddress__email__iexact=login)

        else:
            return None, False

        try:
            user = User.objects.get(lookup_obj)
            return user, is_phone
        except User.DoesNotExist:
            return None, False

    def save(self):
        login = self.cleaned_data.get("login", None)
        if login is not None:
            user, is_phone = self.get_users_and_method(login)

            if user:
                url = reverse(
                    "phone_auth:phone_password_reset_confirm",
                    kwargs={
                        "uidb64": urlsafe_base64_encode(force_bytes(user.pk)),
                        "token": default_token_generator.make_token(user),
                    },
                )
                if is_phone:
                    reset_password_phone.send(
                        sender=self.__class__, user=user, url=url, phone=login
                    )
                else:
                    reset_password_email.send(
                        sender=self.__class__, user=user, url=url, email=login
                    )


class PhoneEmailVerificationForm(forms.Form):
    """
    Send verify_email/verify_phone signal with user instance
    and relative_url which is one-time use link to verify email/phone
    user requested to verify.
    """

    method = forms.CharField(max_length=5)
    pk = forms.IntegerField()

    def save(self, user):
        method = self.cleaned_data.get("method")
        pk = self.cleaned_data.get("pk")
        if method == "email":
            try:
                email_obj = EmailAddress.objects.get(user=user, pk=pk)
                if email_obj.is_verified:
                    return "Email already Verified"

                url = self._get_token_url(
                    email_obj=email_obj, phone_obj=None, user=user
                )
                verify_email.send(
                    sender=self.__class__,
                    user=user,
                    url=url,
                    email=email_obj.email,
                )
                return "Email Verification Sent"
            except EmailAddress.DoesNotExist:
                # In this case say email sent successfully
                # to avoid user enumeration attack
                pass
        elif method == "phone":
            try:
                phone_obj = PhoneNumber.objects.get(user=user, pk=pk)
                if phone_obj.is_verified:
                    return "Phone already Verified"

                url = self._get_token_url(
                    email_obj=None, phone_obj=phone_obj, user=user
                )
                verify_phone.send(
                    sender=self.__class__,
                    user=user,
                    url=url,
                    phone=phone_obj.phone.__str__(),
                )
                return "Phone Verification Sent"
            except PhoneNumber.DoesNotExist:
                pass

        return "Something Went Wrong"

    def _get_token_url(self, email_obj, phone_obj, user):
        token = phone_token_generator(
            email_address_obj=email_obj, phone_number_obj=phone_obj
        ).make_token(user)
        url = reverse(
            "phone_auth:phone_email_verification_confirm",
            kwargs={
                "idb64": self._get_email_phone_b64(email_obj, phone_obj),
                "token": token,
            },
        )
        return url

    @staticmethod
    def _get_email_phone_b64(email_obj, phone_obj):
        if email_obj is not None:
            return urlsafe_base64_encode(force_bytes(f"email{email_obj.pk}"))
        if phone_obj is not None:
            return urlsafe_base64_encode(force_bytes(f"phone{phone_obj.pk}"))


class PhoneLogoutForm(forms.Form):
    pass


class AddPhoneForm(forms.Form):
    """Form to add new phone number"""

    phone = PhoneNumberField(required=True)

    def save(self, user):
        try:
            phone = self.cleaned_data.get("phone")
            PhoneNumber.objects.create(user=user, phone=phone)

        except IntegrityError as e:
            if "UNIQUE constraint" in e.args[0]:
                self.add_error("phone", "Phone already exists")


class AddEmailForm(forms.Form):
    """Form to add new email"""

    email = forms.EmailField(required=True)

    def save(self, user):
        try:
            email = self.cleaned_data.get("email")
            EmailAddress.objects.create(user=user, email=email)

        except IntegrityError as e:
            if "UNIQUE constraint" in e.args[0]:
                self.add_error("email", "Email already exists")



from django.contrib.auth import get_user_model
from django.contrib.auth.backends import ModelBackend
from django.db.models import Q

from . import app_settings
from .app_settings import AuthenticationMethod
from .forms import EmailValidationForm, PhoneValidationForm, UsernameValidationForm

User = get_user_model()


class CustomAuthBackend(ModelBackend):
    def authenticate(self, request, **kwargs):
        """Authenticate with phone/email/username and password """

        login = kwargs.get("login", kwargs.get("username", None))
        password = kwargs.get("password", None)

        if login and password:

            lookup_obj = Q()

            authentication_methods = app_settings.AUTHENTICATION_METHODS
            if (
                AuthenticationMethod.PHONE in authentication_methods
                and PhoneValidationForm({"phone": login}).is_valid()
            ):
                lookup_obj |= Q(phonenumber__phone=login)
            elif (
                AuthenticationMethod.EMAIL in authentication_methods
                and EmailValidationForm({"email": login}).is_valid()
            ):
                lookup_obj |= Q(emailaddress__email__iexact=login)
            elif (
                AuthenticationMethod.USERNAME in authentication_methods
                and UsernameValidationForm({"username": login}).is_valid()
            ):
                lookup_obj |= Q(username__exact=login)
            else:
                return None

            if lookup_obj:
                try:
                    user = User.objects.get(lookup_obj)
                    if user.check_password(password) and self.user_can_authenticate(
                        user
                    ):
                        return user
                except User.DoesNotExist:
                    return None

        return None



import re

from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _


def get_username_regex():
    """Username regex"""

    return r"^[a-zA-Z0-9_.-]{1,150}$"


def validate_username(username):
    if re.search(get_username_regex(), username) is None:
        raise ValidationError(
            _(
                "Username should be 150 characters or fewer."
                "Letters, digits and ./-/_ only."
            ),
            params={"username": username},
        )



from django.contrib.auth.mixins import AccessMixin
from django.shortcuts import redirect

from . import app_settings


class AnonymousRequiredMixin(AccessMixin):
    """Verify that the current user is anonymous."""

    def dispatch(self, request, *args, **kwargs):
        if request.user.is_anonymous:
            return super().dispatch(request, *args, **kwargs)
        return redirect(app_settings.LOGIN_REDIRECT_URL)


class VerifiedPhoneRequiredMixin(AccessMixin):
    """Verify that the user has verified phone."""

    def dispatch(self, request, *args, **kwargs):
        if (
            request.user.is_authenticated
            and request.user.phonenumber_set.filter(is_verified=True).exists()
        ):
            return super().dispatch(request, *args, **kwargs)
        return redirect("phone_auth:phone_email_verification")


class VerifiedEmailRequiredMixin(AccessMixin):
    """Verify that the user has verified email."""

    def dispatch(self, request, *args, **kwargs):
        if (
            request.user.is_authenticated
            and request.user.emailaddress_set.filter(is_verified=True).exists()
        ):
            return super().dispatch(request, *args, **kwargs)
        return redirect("phone_auth:phone_email_verification")



# Generated by Django 3.1.7 on 2021-04-13 15:58

import django.db.models.deletion
import phonenumber_field.modelfields
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="PhoneNumber",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "phone",
                    phonenumber_field.modelfields.PhoneNumberField(
                        max_length=128, region=None, unique=True
                    ),
                ),
                ("is_verified", models.BooleanField(default=False)),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="EmailAddress",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("email", models.EmailField(max_length=254, unique=True)),
                ("is_verified", models.BooleanField(default=False)),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
    ]



from django.contrib import admin

from .models import EmailAddress, PhoneNumber


@admin.register(PhoneNumber)
class PhoneNumberAdmin(admin.ModelAdmin):
    raw_id_fields = ("user",)


@admin.register(EmailAddress)
class EmailAddressAdmin(admin.ModelAdmin):
    raw_id_fields = ("user",)



from django.contrib.auth import REDIRECT_FIELD_NAME
from django.contrib.auth.decorators import login_required, user_passes_test
from django.shortcuts import redirect

from . import app_settings


def anonymous_required(function=None, redirect_url=app_settings.LOGIN_REDIRECT_URL):
    """Decorator to check if a user is anonymous"""

    if not redirect_url:
        redirect_url = "/"

    actual_decorator = user_passes_test(
        lambda u: u.is_anonymous, login_url=redirect_url
    )

    if function:
        return actual_decorator(function)
    return actual_decorator


def verified_email_required(
    function=None, login_url=None, redirect_field_name=REDIRECT_FIELD_NAME
):
    """Decorator to check if the user has verified email"""

    def decorator(view_func):
        @login_required(redirect_field_name=redirect_field_name, login_url=login_url)
        def _wrapped_view(request, *args, **kwargs):
            if (
                request.user.is_authenticated
                and request.user.emailaddress_set.filter(is_verified=True).exists()
            ):
                return view_func(request, *args, **kwargs)
            else:
                return redirect("phone_auth:phone_email_verification")

        return _wrapped_view

    if function:
        return decorator(function)
    return decorator


def verified_phone_required(
    function=None, login_url=None, redirect_field_name=REDIRECT_FIELD_NAME
):
    """Decorator to check if the user has verified phone"""

    def decorator(view_func):
        @login_required(redirect_field_name=redirect_field_name, login_url=login_url)
        def _wrapped_view(request, *args, **kwargs):
            if (
                request.user.is_authenticated
                and request.user.phonenumber_set.filter(is_verified=True).exists()
            ):
                return view_func(request, *args, **kwargs)
            else:
                return redirect("phone_auth:phone_email_verification")

        return _wrapped_view

    if function:
        return decorator(function)
    return decorator



from django.contrib.auth.tokens import PasswordResetTokenGenerator


class PhoneEmailVerificationTokenGenerator(PasswordResetTokenGenerator):
    """Generate/Verify one-time use token for email/phone verification"""

    def __init__(self, email_address_obj, phone_number_obj):
        super(PhoneEmailVerificationTokenGenerator, self).__init__()

        if email_address_obj is None and phone_number_obj is None:
            raise TypeError(
                "Both email_address_obj and" " phone_number_obj arguments can't be None"
            )

        self.email_address_obj = email_address_obj
        self.phone_number_obj = phone_number_obj

    def _make_hash_value(self, user, timestamp):

        if self.email_address_obj:
            email = self.email_address_obj.email
            is_verified = self.email_address_obj.is_verified
            return f"{user.pk}{timestamp}{email}{is_verified}"

        elif self.phone_number_obj:
            phone = self.phone_number_obj.phone
            country_code = phone.country_code
            national_number = phone.national_number
            is_verified = self.phone_number_obj.is_verified
            return f"{user.pk}{timestamp}{country_code}{national_number}{is_verified}"


phone_token_generator = PhoneEmailVerificationTokenGenerator



from pathlib import Path

BASE_DIR = Path(__file__).resolve().parent.parent
SECRET_KEY = 'n@8D*&MS38$*%S4t_YtQyLUew&*&m_zn8$*F@'
DEBUG = True

ALLOWED_HOSTS = ['*']
INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'phonenumber_field',
    'phone_auth',
    'debug_toolbar',
]

MIDDLEWARE = [
    'debug_toolbar.middleware.DebugToolbarMiddleware',
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'django_phone_auth_project.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

WSGI_APPLICATION = 'django_phone_auth_project.wsgi.application'
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': BASE_DIR / 'db.sqlite3',
    }
}

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]

AUTHENTICATION_BACKENDS = [
    'phone_auth.backend.CustomAuthBackend',
    'django.contrib.auth.backends.ModelBackend',
]

LANGUAGE_CODE = 'en-us'
TIME_ZONE = 'UTC'
USE_I18N = True
USE_L10N = True
USE_TZ = True

STATIC_URL = '/static/'

# django-debug-toolbar settings
INTERNAL_IPS = [
    '127.0.0.1',
]
